use crate::{Meta, Relationships, Links, GenericObject, ResourceObject, Attributes};
use serde::{Serialize as SerializeTrait, de::DeserializeOwned};
use serde_derive::{Serialize, Deserialize};
use std::convert::{TryFrom, TryInto};
use super::ObjectConversionError;

/// A resource identifier object
///
/// Basically just the identifying information needed to refer to a `ResourceObject`
///
/// See the [JSON:API docs](https://jsonapi.org/format/#document-resource-identifier-objects)
/// for more information
#[derive(Serialize, Deserialize, Debug, PartialEq, Clone)]
pub struct Identifier {
    /// The identifier of the object which, along with the object's type, MUST refer to a single
    /// unique resource
    pub id: String,
    /// The object's type, renamed here to kind due to keyword limitations
    #[serde(rename = "type")]
    pub kind: String,
    /// Non-standard meta information
    #[serde(skip_serializing_if = "Option::is_none")]
    pub meta: Option<Meta>,
}

impl Identifier {
    pub fn new(id: String, kind: String) -> Self {
        Self {
            id: id,
            kind,
            meta: None,
        }
    }
}

impl TryFrom<GenericObject> for Identifier {
    type Error = ObjectConversionError;
    fn try_from(mut go: GenericObject) -> Result<Self, Self::Error> {
        go.id.take().map(|id| 
            Self {
                id,
                kind: go.kind,
                meta: go.meta,
            }).ok_or(ObjectConversionError::MissingId)
    }
}

impl TryFrom<&GenericObject> for Identifier {
    type Error = ObjectConversionError;
    fn try_from(go: &GenericObject) -> Result<Self, Self::Error> {
        go.id.as_ref().map(|id| 
            Self {
                id: id.clone(),
                kind: go.kind.clone(),
                meta: go.meta.clone(),
            }).ok_or(ObjectConversionError::MissingId)
    }
}

impl<A> TryFrom<ResourceObject<A>> for Identifier 
where A: Attributes + SerializeTrait + DeserializeOwned {
    type Error = ObjectConversionError;
    fn try_from(mut ro: ResourceObject<A>) -> Result<Self, Self::Error> {
        ro.id.take().map(|id| 
            Self {
                id,
                kind: A::kind(),
                meta: ro.meta,
            }).ok_or(ObjectConversionError::MissingId)
    }
}

impl<A> TryFrom<&ResourceObject<A>> for Identifier
where A: Attributes + SerializeTrait + DeserializeOwned {
    type Error = ObjectConversionError;
    fn try_from(ro: &ResourceObject<A>) -> Result<Self, Self::Error> {
        ro.id.as_ref().map(|id| 
            Self {
                id: id.clone(),
                kind: A::kind(),
                meta: ro.meta.clone(),
            }).ok_or(ObjectConversionError::MissingId)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use serde_json;

    #[test]
    fn serde() {
        let id1 = Identifier{
            id: "a".into(),
            kind: "b".into(),
            meta: None,
        };
        let s = serde_json::to_string(&id1).unwrap();
        assert_eq!(s, "{\"id\":\"a\",\"type\":\"b\"}");
        let id2 = serde_json::from_str(&s).unwrap();
        assert_eq!(id1, id2);
    }

    #[test]
    fn from_go() {
        let mut meta = Meta::new();
        meta.insert("c".into(), serde_json::json!("d"));
        let go = GenericObject {
            id: Some("a".into()),
            kind: "b".into(),
            attributes: None,
            relationships: None,
            links: None,
            meta: Some(meta.clone()),
        };

        let id = (&go).try_into().unwrap();
        assert_eq!(Identifier {
            id: "a".into(),
            kind: "b".into(),
            meta: Some(meta.clone()),
        }, id);

        let id = go.try_into().unwrap();
        assert_eq!(Identifier {
            id: "a".into(),
            kind: "b".into(),
            meta: Some(meta.clone()),
        }, id);
    }

    #[test] 
    fn from_ro() {
        #[derive(Serialize, Deserialize, Eq, PartialEq, Clone)]
        struct Attr;
        impl Attributes for Attr{
            fn kind() -> String { "b".into() }
        }

        let mut meta = Meta::new();
        meta.insert("c".into(), serde_json::json!("d"));
        let ro : ResourceObject<Attr> = ResourceObject {
            id: Some("a".into()),
            attributes: None,
            relationships: None,
            links: None,
            meta: Some(meta.clone()),
        };

        let id = (&ro).try_into().unwrap();
        assert_eq!(Identifier {
            id: "a".into(),
            kind: "b".into(),
            meta: Some(meta.clone()),
        }, id);

        let id = ro.try_into().unwrap();
        assert_eq!(Identifier {
            id: "a".into(),
            kind: "b".into(),
            meta: Some(meta.clone()),
        }, id);
    }
}
