use serde_derive::{Serialize, Deserialize};
use serde::de::{DeserializeOwned, Deserializer, Deserialize, Error};

/// Represents a value that can either exist, not exist, or be `null`
///
/// This is a helper object for building Attributes. When including it in an
/// attribute use `#[serde(skip_serializing_if = "Optional::is_not_present", default)]`.
#[derive(Serialize, PartialEq, Eq, Debug, Clone)]
#[serde(untagged)]
pub enum Optional<T> {
    /// The key was present in the document and had a value
    Present(T),
    /// The key was not present
    NotPresent, 
    /// The key was present and was null
    Null,
}

impl<T> Optional<T> {
    /// The key was present and had a value
    pub fn is_present(&self) -> bool {
        match self {
            Optional::Present(_) => true,
            _ => false,
        }
    }

    /// The key was not present
    pub fn is_not_present(&self) -> bool {
        match self {
            Optional::NotPresent => true,
            _ => false,
        }
    }

    /// The key was present and was null
    pub fn is_null(&self) -> bool {
        match self {
            Optional::Null => true,
            _ => false,
        }
    }

    /// Map the value if it was present
    pub fn map<U, F>(self, f: F) -> Optional<U> where F: FnOnce(T) -> U {
        match self {
            Optional::Present(t) => Optional::Present(f(t)),
            Optional::NotPresent => Optional::NotPresent,
            Optional::Null => Optional::Null,
        }
    }

    /// Maps `Present` to `Some`, calls `f` is `Null` and maps the result to `Some`,
    /// maps `NotPresent` to `None`
    pub fn map_null<F>(self, f: F) -> Option<T> where F: FnOnce() -> T {
        match self {
            Optional::Present(t) => Some(t),
            Optional::NotPresent => None,
            Optional::Null => Some(f()),
        }
    }

    /// Converts to `Some` if `Present`, otherwise converts to `None`
    pub fn as_option(self) -> Option<T> {
        match self {
            Optional::Present(t) => Some(t),
            _ => None,
        }
    }

    /// Converts from `&Optional<T>` to `Optional<&T>`
    pub fn as_ref(&self) -> Optional<&T> {
        match self {
            Optional::Present(t) => Optional::Present(&t), 
            Optional::NotPresent => Optional::NotPresent,
            Optional::Null => Optional::Null,
        }
    }

    /// Converts from `&mut Optional<T>` to `Optional<&mut T>`
    pub fn as_mut(&mut self) -> Optional<&mut T> {
        match self {
            Optional::Present(t) => Optional::Present(t), 
            Optional::NotPresent => Optional::NotPresent,
            Optional::Null => Optional::Null,
        }
    }

    /// Unwraps an Optional, yeilding the content of a `Present`
    ///
    /// # Panics
    /// Panics if the value is not a `Present` with a custom panic message provided by `msg`
    pub fn expect(self, msg: &str) -> T {
        match self {
            Optional::Present(t) => t,
            _ => panic!("{}", msg),
        }
    }

    /// Move the value `v` out of the `Optional<T>` if it is `Present(v)`.
    ///
    /// # Panics
    /// Panics if the value is not a `Present`
    pub fn unwrap(self) -> T {
        match self {
            Optional::Present(t) => t,
            Optional::NotPresent => panic!("Called unwrap on a NotPresent value"),
            Optional::Null => panic!("Called unwrap on a Null value"),
        }
    }

    /// Unwraps if `Present`, returns `null` if `Null`
    ///
    /// # Panics
    /// Panics if the value is `NotPresent`
    pub fn unwrap_or_null(self, null: T) -> T {
        match self {
            Optional::Present(t) => t,
            Optional::NotPresent => panic!("Called unwrap on a NotPresent value"),
            Optional::Null => null,
        }
    }

    /// Unwraps if `Present`, calls `null` and returns the result if `Null`
    ///
    /// # Panics
    /// Panics if the value is `NotPresent`
    pub fn unwrap_or_else_null<F>(self, null: F) -> T where F: FnOnce() -> T {
        match self {
            Optional::Present(t) => t,
            Optional::NotPresent => panic!("Called unwrap on a NotPresent value"),
            Optional::Null => null(),
        }
    }
}

/// Defaults to `NotPresent`
impl<T> Default for Optional<T> {
    fn default() -> Self {
        Optional::NotPresent
    }
}

/// Maps `None` to `NotPresent`
impl<T> From<Option<T>> for Optional<T> {
    fn from(o: Option<T>) -> Self {
        match o {
            Some(t) => Optional::Present(t),
            None => Optional::NotPresent,
        }
    }
}

impl<'de, T> Deserialize<'de> for Optional<T>
where T: DeserializeOwned {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where D: Deserializer<'de> {
        Ok(match Option::<T>::deserialize(deserializer)? {
            Some(t) => Optional::Present(t),
            None => Optional::Null,
        })
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[derive(Serialize, Deserialize, PartialEq, Eq, Debug)]
    struct TestStruct {
        #[serde(skip_serializing_if = "Optional::is_not_present", default)]
        inner: Optional<usize>,
    }

    #[test]
    fn serde_present() {
        let ts1 = TestStruct{ inner: Optional::Present(1), };
        let s = serde_json::to_string(&ts1).unwrap();
        assert_eq!(s, "{\"inner\":1}");
        let ts2 = serde_json::from_str(&s).unwrap();
        assert_eq!(ts1, ts2);
    }

    #[test]
    fn serde_not_present() {
        let ts1 = TestStruct{ inner: Optional::NotPresent, };
        let s = serde_json::to_string(&ts1).unwrap();
        assert_eq!(s, "{}");
        let ts2 = serde_json::from_str(&s).unwrap();
        assert_eq!(ts1, ts2);
    }

    #[test]
    fn serde_null() {
        let ts1 = TestStruct{ inner: Optional::Null, };
        let s = serde_json::to_string(&ts1).unwrap();
        assert_eq!(s, "{\"inner\":null}");
        let ts2 = serde_json::from_str(&s).unwrap();
        assert_eq!(ts1, ts2);
    }

    #[test]
    #[should_panic]
    fn failed_deserialize() {
        let _ : TestStruct = serde_json::from_str("{\"inner\":\"oops i'm a string\"}").unwrap();
    }

    #[test]
    fn default() {
        assert_eq!(Optional::<usize>::default(), Optional::NotPresent);
    }

    #[test]
    fn is() {
        let o : Optional<usize> = Optional::Present(1);
        assert!(o.is_present());
        assert!(!o.is_not_present());
        assert!(!o.is_null());

        let o : Optional<usize> = Optional::NotPresent;
        assert!(!o.is_present());
        assert!(o.is_not_present());
        assert!(!o.is_null());

        let o : Optional<usize> = Optional::Null;
        assert!(!o.is_present());
        assert!(!o.is_not_present());
        assert!(o.is_null());
    }

    #[test]
    fn map() {
        let o: Optional<usize> = Optional::Present(1);
        assert_eq!(o.map(|i| i.to_string()), Optional::Present("1".into()));

        let o: Optional<usize> = Optional::NotPresent;
        assert_eq!(o.map(|i| i.to_string()), Optional::NotPresent);

        let o: Optional<usize> = Optional::Null;
        assert_eq!(o.map(|i| i.to_string()), Optional::Null);
    }

    #[test]
    fn map_null() {
        let o: Optional<usize> = Optional::Present(1);
        assert_eq!(o.map_null(|| 2), Some(1));

        let o: Optional<usize> = Optional::NotPresent;
        assert_eq!(o.map_null(|| 2), None);

        let o: Optional<usize> = Optional::Null;
        assert_eq!(o.map_null(|| 2), Some(2));
    }

    #[test]
    fn test_as() {
        let mut o: Optional<usize> = Optional::Present(1);
        assert_eq!(o.as_ref(), Optional::Present(&1));
        assert_eq!(o.as_mut(), Optional::Present(&mut 1));
        assert_eq!(o.as_option(), Some(1));

        let mut o: Optional<usize> = Optional::NotPresent;
        assert_eq!(o.as_ref(), Optional::NotPresent);
        assert_eq!(o.as_mut(), Optional::NotPresent);
        assert_eq!(o.as_option(), None);

        let mut o: Optional<usize> = Optional::Null;
        assert_eq!(o.as_ref(), Optional::Null);
        assert_eq!(o.as_mut(), Optional::Null);
        assert_eq!(o.as_option(), None);
    }

    #[test]
    fn unwrap_success() {
        let o: Optional<usize> = Optional::Present(1);
        assert_eq!(o.expect("test"), 1);

        let o: Optional<usize> = Optional::Present(1);
        assert_eq!(o.unwrap(), 1);

        let o: Optional<usize> = Optional::Present(1);
        assert_eq!(o.unwrap_or_null(2), 1);

        let o: Optional<usize> = Optional::Null;
        assert_eq!(o.unwrap_or_null(2), 2);

        let o: Optional<usize> = Optional::Present(1);
        assert_eq!(o.unwrap_or_else_null(|| 2), 1);

        let o: Optional<usize> = Optional::Null;
        assert_eq!(o.unwrap_or_else_null(|| 2), 2);
    }

    #[test]
    #[should_panic]
    fn expect_not_present() {
        let o: Optional<usize> = Optional::NotPresent;
        o.expect("test");
    }

    #[test]
    #[should_panic]
    fn expect_null() {
        let o: Optional<usize> = Optional::Null;
        o.expect("test");
    }

    #[test]
    #[should_panic]
    fn unwrap_not_present() {
        let o: Optional<usize> = Optional::NotPresent;
        o.unwrap();
    }

    #[test]
    #[should_panic]
    fn unwrap_null() {
        let o: Optional<usize> = Optional::Null;
        o.unwrap();
    }

    #[test]
    #[should_panic]
    fn unwrap_or_null() {
        let o: Optional<usize> = Optional::NotPresent;
        o.unwrap_or_null(2);
    }

    #[test]
    #[should_panic]
    fn unwrap_or_else_null() {
        let o: Optional<usize> = Optional::NotPresent;
        o.unwrap_or_else_null(|| 2);
    }
}
