use serde_derive::{Serialize, Deserialize};
use serde::de::{DeserializeOwned, Deserializer, Deserialize, Error};
use serde_json::{Value, self};

/// A rather complicated relative of `Option<Vec>`
///
/// There are a couple places in JSON:API where you have keys with values that can either:
/// - Not Exist
/// - Be `null`
/// - Contain a single object
/// - Contain an array of objects
///
/// This enum exists for those cases
///
/// `OptionalVec::NotPresent` exists to specify that the key should not (or did not)
/// exist
///
/// `OptionalVec::One` is `None` when the value is `null` and `Some` when there is a 
/// single object
///
/// `OptionalVec::Many` will be used when the value is an array, empty or not
#[derive(Serialize, PartialEq, Eq, Debug, Clone)]
#[serde(untagged)]
pub enum OptionalVec<T> {
    NotPresent,
    One(Option<T>),
    Many(Vec<T>),
}

impl<T> OptionalVec<T> {
    /// Checks for the `OptionalVec::NotPresent` variant
    ///
    /// Mostly provided to allow serialization to be skipped
    pub fn is_not_present(&self) -> bool {
        match self {
            OptionalVec::NotPresent => true,
            _ => false,
        }
    }

    /// Checks for the `OptionalVec::One` variant
    pub fn is_one(&self) -> bool {
        match self {
            OptionalVec::One(_) => true,
            _ => false,
        }
    }

    /// Checks for the `OptionalVec::Many` variant
    pub fn is_many(&self) -> bool {
        match self {
            OptionalVec::Many(_) => true,
            _ => false,
        }
    }

    /// Map each contained value with the given function
    pub fn map<U, F>(self, f: F) -> OptionalVec<U> where F: Fn(T) -> U {
        match self {
            OptionalVec::NotPresent => OptionalVec::NotPresent,
            OptionalVec::One(Some(t)) => OptionalVec::One(Some(f(t))),
            OptionalVec::One(None) => OptionalVec::One(None),
            OptionalVec::Many(v) => OptionalVec::Many(v.into_iter().map(f).collect()),
        }
    }

    /// Error with `e` if `Many`, else build an `Option`
    pub fn one_or<E>(self, e: E) -> Result<Option<T>, E> {
        match self {
            OptionalVec::Many(_) => Err(e),
            OptionalVec::One(Some(t)) => Ok(Some(t)),
            _ => Ok(None),
        }
    }

    /// Return the contents of `Many`, else error with `e`
    pub fn many_or<E>(self, e: E) -> Result<Vec<T>, E> {
        match self {
            OptionalVec::Many(v) => Ok(v),
            _ => Err(e),
        }
    }

    /// Return all contained items as a Vec
    ///
    /// Will map `OptionalVec::One(Some(v))` into `vec![v]`
    pub fn into_vec(self) -> Vec<T> {
        match self {
            OptionalVec::Many(v) => v,
            OptionalVec::One(Some(i)) => vec![i],
            _ => vec![],
        }
    }

    /// Unwraps `One(Some(v))`
    ///
    /// # Panics
    /// Panics for all other variants
    pub fn unwrap_one(self) -> T {
        match self {
            OptionalVec::One(Some(v)) => v,
            OptionalVec::Many(_) => panic!("OptionalVec expected one got many"),
            _ => panic!("OptionalVec expected one got none"),
        }
    }
    
    /// Unwraps `Many(v)`
    ///
    /// # Panics
    /// Panics for all other variants
    pub fn unwrap_many(self) -> Vec<T> {
        match self {
            OptionalVec::Many(v) => v,
            OptionalVec::One(Some(_)) => panic!("OptionalVec expected many got one"),
            _ => panic!("OptionalVec expected many got none"),
        }
    }
}

impl<'de, T> Deserialize<'de> for OptionalVec<T> 
where T: DeserializeOwned {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error> 
    where D: Deserializer<'de> {
        // okay so this should probably use a visitor but that's like
        //
        // effort
        //
        // ya know?
        let v = Value::deserialize(deserializer)?;
        match serde_json::from_value::<Option<T>>(v.clone()) {
            Ok(one) => Ok(OptionalVec::One(one)),
            Err(_) => match serde_json::from_value(v) {
                Ok(many) => Ok(OptionalVec::Many(many)),
                Err(_) => Err(D::Error::custom("Neither one nor many")),
            },
        }
    }
}

/// Defaults to `OptionalVec::NotPresent`
impl<T> Default for OptionalVec<T> {
    fn default() -> Self {
        OptionalVec::NotPresent
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[derive(Serialize, Deserialize, PartialEq, Eq, Debug)]
    struct TestStruct {
        #[serde(skip_serializing_if = "OptionalVec::is_not_present", default)]
        inner: OptionalVec<usize>,
    }

    #[test]
    fn serde_not_present() {
        let ts1 = TestStruct{ inner: OptionalVec::NotPresent };
        let s = serde_json::to_string(&ts1).unwrap();
        assert_eq!(s, "{}");
        let ts2 = serde_json::from_str(&s).unwrap();
        assert_eq!(ts1, ts2);
    }

    #[test]
    fn serde_empty_to_one() {
        let ts1 = TestStruct{ inner: OptionalVec::One(None) };
        let s = serde_json::to_string(&ts1).unwrap();
        assert_eq!(s, "{\"inner\":null}");
        let ts2 = serde_json::from_str(&s).unwrap();
        assert_eq!(ts1, ts2);
    }

    #[test]
    fn serde_to_one() {
        let ts1 = TestStruct{ inner: OptionalVec::One(Some(1)) };
        let s = serde_json::to_string(&ts1).unwrap();
        assert_eq!(s, "{\"inner\":1}");
        let ts2 = serde_json::from_str(&s).unwrap();
        assert_eq!(ts1, ts2);
    }

    #[test]
    fn serde_to_many() {
        let ts1 = TestStruct{ inner: OptionalVec::Many(vec![1,2]) };
        let s = serde_json::to_string(&ts1).unwrap();
        assert_eq!(s, "{\"inner\":[1,2]}");
        let ts2 = serde_json::from_str(&s).unwrap();
        assert_eq!(ts1, ts2);
    }

    #[test]
    #[should_panic]
    fn failed_deserialize() {
        let _ : TestStruct = serde_json::from_str("{\"inner\":\"oops i'm a string\"}").unwrap();
    }

    #[test]
    fn is() {
        let ov : OptionalVec<usize> = OptionalVec::NotPresent;
        assert!(ov.is_not_present());
        assert!(!ov.is_one());
        assert!(!ov.is_many());

        let ov : OptionalVec<usize> = OptionalVec::One(None);
        assert!(!ov.is_not_present());
        assert!(ov.is_one());
        assert!(!ov.is_many());

        let ov : OptionalVec<usize> = OptionalVec::Many(vec![]);
        assert!(!ov.is_not_present());
        assert!(!ov.is_one());
        assert!(ov.is_many());
    }

    #[test]
    fn map() {
        let ov : OptionalVec<usize> = OptionalVec::NotPresent;
        assert_eq!(ov.map(|i| i + 1), OptionalVec::NotPresent);

        let ov : OptionalVec<usize> = OptionalVec::One(Some(1));
        assert_eq!(ov.map(|i| i + 1), OptionalVec::One(Some(2)));

        let ov : OptionalVec<usize> = OptionalVec::One(None);
        assert_eq!(ov.map(|i| i + 1), OptionalVec::One(None));

        let ov : OptionalVec<usize> = OptionalVec::Many(vec![1, 2]);
        assert_eq!(ov.map(|i| i + 1), OptionalVec::Many(vec![2, 3]));
    }

    #[test]
    fn one_or() {
        let ov : OptionalVec<usize> = OptionalVec::NotPresent;
        assert_eq!(ov.one_or("oops"), Ok(None));

        let ov : OptionalVec<usize> = OptionalVec::One(Some(1));
        assert_eq!(ov.one_or("oops"), Ok(Some(1)));

        let ov : OptionalVec<usize> = OptionalVec::One(None);
        assert_eq!(ov.one_or("oops"), Ok(None));

        let ov : OptionalVec<usize> = OptionalVec::Many(vec![1, 2]);
        assert_eq!(ov.one_or("oops"), Err("oops"));
    }

    #[test]
    fn many_or() {
        let ov : OptionalVec<usize> = OptionalVec::NotPresent;
        assert_eq!(ov.many_or("oops"), Err("oops"));

        let ov : OptionalVec<usize> = OptionalVec::One(Some(1));
        assert_eq!(ov.many_or("oops"), Err("oops"));

        let ov : OptionalVec<usize> = OptionalVec::One(None);
        assert_eq!(ov.many_or("oops"), Err("oops"));

        let ov : OptionalVec<usize> = OptionalVec::Many(vec![1, 2]);
        assert_eq!(ov.many_or("oops"), Ok(vec![1, 2]));
    }

    #[test]
    fn into_vec() {
        let ov : OptionalVec<usize> = OptionalVec::NotPresent;
        assert_eq!(ov.into_vec(), Vec::<usize>::new());

        let ov : OptionalVec<usize> = OptionalVec::One(Some(1));
        assert_eq!(ov.into_vec(), vec![1]);

        let ov : OptionalVec<usize> = OptionalVec::One(None);
        assert_eq!(ov.into_vec(), Vec::<usize>::new());

        let ov : OptionalVec<usize> = OptionalVec::Many(vec![1, 2]);
        assert_eq!(ov.into_vec(), vec![1, 2]);
    }

    #[test]
    #[should_panic]
    fn unwrap_one_not_present() {
        let ov : OptionalVec<usize> = OptionalVec::NotPresent;
        ov.unwrap_one();
    }

    #[test]
    #[should_panic]
    fn unwrap_one_none() {
        let ov : OptionalVec<usize> = OptionalVec::One(None);
        ov.unwrap_one();
    }

    #[test]
    fn unwrap_one_some() {
        let ov : OptionalVec<usize> = OptionalVec::One(Some(1));
        assert_eq!(ov.unwrap_one(), 1);
    }

    #[test]
    #[should_panic]
    fn unwrap_one_many() {
        let ov : OptionalVec<usize> = OptionalVec::Many(vec![1, 2]);
        ov.unwrap_one();
    }

    #[test]
    #[should_panic]
    fn unwrap_many_not_present() {
        let ov : OptionalVec<usize> = OptionalVec::NotPresent; 
        ov.unwrap_many();
    }
    
    #[test]
    #[should_panic]
    fn unwrap_many_none() {
        let ov : OptionalVec<usize> = OptionalVec::One(None);
        ov.unwrap_many();
    }
    
    #[test]
    #[should_panic]
    fn unwrap_many_some() {
        let ov : OptionalVec<usize> = OptionalVec::One(Some(1));
        ov.unwrap_many();
    }

    #[test]
    fn unwrap_many_many() {
        let ov : OptionalVec<usize> = OptionalVec::Many(vec![1, 2]);
        assert_eq!(ov.unwrap_many(), vec![1, 2]);
    }
}
